from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

from .errors import GQLPermissionDeniedError
from .settings import gdx_settings
from .typing import ParamSpec, TypeVar

if TYPE_CHECKING:
    from django.db.models import Model

    from .typing import Any, AnyUser, Callable, GraphQLFilterInfo, PermCheck


T = TypeVar("T")
P = ParamSpec("P")


__all__ = [
    "AllowAny",
    "AllowAuthenticated",
    "AllowStaff",
    "AllowSuperuser",
    "BasePermission",
    "restricted_field",
]


class BasePermission:
    @classmethod
    def has_permission(cls, user: AnyUser) -> bool:  # pragma: no cover
        return False

    @classmethod
    def has_node_permission(cls, instance: Model, user: AnyUser, filters: GraphQLFilterInfo) -> bool:
        return cls.has_permission(user)

    @classmethod
    def has_filter_permission(cls, user: AnyUser, filters: GraphQLFilterInfo) -> bool:
        return cls.has_permission(user)

    @classmethod
    def has_mutation_permission(cls, user: AnyUser, input_data: dict[str, Any]) -> bool:
        return cls.has_permission(user)

    @classmethod
    def has_create_permission(cls, user: AnyUser, input_data: dict[str, Any]) -> bool:
        return cls.has_mutation_permission(user, input_data)

    @classmethod
    def has_update_permission(cls, instance: Model, user: AnyUser, input_data: dict[str, Any]) -> bool:
        return cls.has_mutation_permission(user, input_data)

    @classmethod
    def has_delete_permission(cls, instance: Model, user: AnyUser, input_data: dict[str, Any]) -> bool:
        return cls.has_mutation_permission(user, input_data)


class AllowAny(BasePermission):
    @classmethod
    def has_permission(cls, user: AnyUser) -> bool:
        return True


class AllowAuthenticated(BasePermission):
    @classmethod
    def has_permission(cls, user: AnyUser) -> bool:  # pragma: no cover
        return user.is_authenticated


class AllowStaff(BasePermission):
    @classmethod
    def has_permission(cls, user: AnyUser) -> bool:  # pragma: no cover
        return user.is_staff


class AllowSuperuser(BasePermission):
    @classmethod
    def has_permission(cls, user: AnyUser) -> bool:  # pragma: no cover
        return user.is_superuser


def restricted_field(check: PermCheck, *, message: str = "") -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    Decorator for GraphQL field resolvers, which will raise
    a PermissionDenied error if the request user does not have
    appropriate permissions based on the given check.

    :param check: A callable, which takes the request user, and the ObjectType's Model instance
                  (or just the request user) as its arguments.
    :param message: The message to raise in the PermissionError.
                    If not given, use default message from settings.
    """
    message = message or gdx_settings.FIELD_PERMISSION_ERROR_MESSAGE

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                if check(args[1].context.user, args[0]):
                    return func(*args, **kwargs)  # pragma: no cover
            except TypeError:
                if check(args[1].context.user):
                    return func(*args, **kwargs)  # pragma: no cover

            raise GQLPermissionDeniedError(message, gdx_settings.FIELD_PERMISSION_ERROR_CODE)

        return wrapper

    return decorator



from __future__ import annotations

import json
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from inspect import cleandoc
from typing import TYPE_CHECKING

import pytest
import sqlparse
from django import db
from django.contrib.auth import get_user_model
from django.test import Client
from graphene_django.utils.testing import graphql_query

from ..typing import NamedTuple, TypedDict, TypeVar

if TYPE_CHECKING:
    from collections.abc import Generator

    from django.contrib.auth.models import User
    from django.core.files.uploadedfile import SimpleUploadedFile
    from django.http import HttpResponse

    from ..typing import Any, Callable, FieldError, Self

__all__ = [
    "GraphQLClient",
    "capture_database_queries",
    "parametrize_helper",
    "compare_unordered",
]


class GQLResponse:
    def __init__(self, response: HttpResponse, query_data: QueryData) -> None:
        # 'django.test.client.Client.request' sets json attribute on the response.
        self.json: dict[str, Any] = response.json()  # type: ignore[attr-defined]
        self.query_data = query_data

    def __str__(self) -> str:
        return json.dumps(self.json, indent=2, sort_keys=True, default=str)

    def __repr__(self) -> str:
        return repr(self.json)

    def __len__(self) -> int:
        return len(self.edges)

    def __getitem__(self, item: str) -> dict[str, Any] | None:
        return (self.data or {})[item]

    def __contains__(self, item: str) -> bool:
        return item not in self.json

    @property
    def queries(self) -> list[str]:
        """Return a list of the database queries that were executed."""
        return self.query_data.queries

    @property
    def query_log(self) -> str:
        """Return a string representation of the database queries that were executed."""
        return self.query_data.log

    @property
    def data(self) -> dict[str, Any] | None:
        """Return the data from the response content."""
        return self.json["data"]

    @property
    def first_query_object(self) -> dict[str, Any] | list[Any] | None:
        """
        Return the first query object in the response content.

        >>> self.json = {"data": {"foo": {"name": "bar"}}}
        >>> self.first_query_object
        {"name": "bar"}
        """
        data = self.data or {}
        try:
            return next(iter(data.values()))
        except StopIteration:  # pragma: no cover
            msg = f"No query object not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    @property
    def edges(self) -> list[dict[str, Any]]:
        """
        Return edges from the first query in the response content.

        >>> self.json = {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.edges
        [{"node": {"name": "bar"}}]
        """
        try:
            return self.first_query_object["edges"]
        except (KeyError, TypeError):  # pragma: no cover
            msg = f"Edges not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    def node(self, index: int = 0) -> dict[str, Any]:
        """
        Return the node at the given index in the response content edges.

        >>> self.json = {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.node(0)
        {"name": "bar"}
        """
        try:
            return self.edges[index]["node"]
        except (IndexError, TypeError):  # pragma: no cover
            msg = f"Node {index!r} not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    @property
    def has_errors(self) -> bool:  # pragma: no cover
        """Are there any errors in the response?"""
        return "errors" in self.json and self.json.get("errors") is not None

    @property
    def errors(self) -> list[dict[str, Any]]:
        """
        Return all errors.

        >>> self.json = {"errors": [{"message": "bar", "path": [...], ...}]}
        >>> self.errors
        [{"message": "bar", "path": [...], ...}]
        """
        try:
            return self.json["errors"]
        except (KeyError, TypeError):  # pragma: no cover
            msg = f"Errors not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    def error_message(self, selector: int | str = 0) -> str:  # pragma: no cover
        """
        Return the error message from the errors list...

        1) in the given index

        >>> self.json = {"errors": [{"message": "bar", ...}]}
        >>> self.error_message(0)  # default
        "bar"

        2) in the given path:

        >>> self.json = {"errors": [{"message": "bar", "path": ["fizz", "buzz", "foo"], ...}]}
        >>> self.error_message("foo")
        "bar"
        """
        if isinstance(selector, int):
            try:
                return self.errors[selector]["message"]
            except IndexError:
                msg = f"Errors list doesn't have an index {selector}: {self.json}"
                pytest.fail(msg, pytrace=False)
            except (KeyError, TypeError):
                msg = f"Field 'message' not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)
        else:
            try:
                return next(error["message"] for error in self.errors if error["path"][-1] == selector)
            except StopIteration:
                msg = f"Errors list doesn't have an error for field '{selector}': {self.json}"
                pytest.fail(msg, pytrace=False)
            except (KeyError, TypeError):
                msg = f"Field 'message' not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)

    @property
    def field_errors(self) -> list[FieldError]:
        """
        Return all field error messages.

        >>> self.json = {
        ...     "errors": [
        ...         {
        ...             "extensions": {
        ...                 "errors": [
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": ["bar"],
        ...                         "codes": ["baz"],
        ...                     },
        ...                 ],
        ...             },
        ...         },
        ...     ],
        ... }
        ...
        >>> self.field_errors
        [{"field": "foo", "messages": ["bar"], "codes": ["baz"]}]
        """
        try:
            return [error for item in self.errors for error in item.get("extensions", {}).get("errors", [])]
        except (KeyError, TypeError):  # pragma: no cover
            msg = f"Field errors not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    def field_error_messages(self, field: str = "nonFieldErrors") -> list[str]:  # pragma: no cover
        """
        Return field error messages for desired field.

        >>> self.json = {
        ...     "errors": [
        ...         {
        ...             "extensions": {
        ...                 "errors": [
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "bar",
        ...                         "codes": "baz",
        ...                     },
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "one",
        ...                         "codes": "",
        ...                     },
        ...                 ],
        ...             },
        ...         },
        ...     ],
        ... }
        ...
        >>> self.field_error_messages("foo")
        ["bar", "one"]
        """
        messages: list[str] = []
        for error in self.field_errors:
            if error.get("field") == field:
                try:
                    messages.append(error["message"])
                except (KeyError, TypeError):
                    msg = f"Error message for field {field!r} not found in error: {error}"
                    pytest.fail(msg, pytrace=False)

        return messages

    def error_code(self, selector: int | str = 0) -> str:  # pragma: no cover
        """
        Return the error code from the errors list.

        1) in the given index

        >>> self.json = {"errors": [{"extensions": {"code": "bar"}, ...}]}
        >>> self.error_code(0)  # default
        "bar"

        2) in the given path:

        >>> self.json = {"errors": [{"extensions": {"code": "bar"}, "path": ["fizz", "buzz", "foo"], ...}]}
        >>> self.error_code("foo")
        "bar"
        """
        if isinstance(selector, int):
            try:
                return self.errors[selector]["extensions"]["code"]
            except (KeyError, TypeError):
                msg = f"Error code not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)
        else:
            try:
                return next(error["extensions"]["code"] for error in self.errors if error["path"][-1] == selector)
            except StopIteration:
                msg = f"Errors list doesn't have an error for field '{selector}': {self.json}"
                pytest.fail(msg, pytrace=False)
            except (KeyError, TypeError):
                msg = f"Field 'extensions' not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)

    def field_error_codes(self, field: str = "nonFieldErrors") -> list[str]:  # pragma: no cover
        """
        Return field error codes for desired field.

        >>> self.json = {
        ...     "errors": [
        ...         {
        ...             "extensions": {
        ...                 "errors": [
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "bar",
        ...                         "codes": "baz",
        ...                     },
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "one",
        ...                         "codes": "",
        ...                     },
        ...                 ],
        ...             },
        ...         },
        ...     ],
        ... }
        ...
        >>> self.field_error_codes("foo")
        ["bar", ""]
        """
        codes: list[str] = []
        for error in self.field_errors:
            if error.get("field") == field:
                try:
                    codes.append(error["code"])
                except (KeyError, TypeError):
                    msg = f"Error code for field {field!r} not found in error: {error}"
                    pytest.fail(msg, pytrace=False)

        return codes

    def assert_query_count(self, count: int) -> None:  # pragma: no cover
        if len(self.queries) != count:
            msg = f"Expected {count} queries, got {len(self.queries)}.\n{self.query_log}"
            pytest.fail(msg, pytrace=False)


class GraphQLClient(Client):
    def __call__(  # noqa: PLR0913
        self: Self,
        query: str,
        operation_name: str | None = None,
        input_data: dict[str, Any] | None = None,
        variables: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        files: dict[str, SimpleUploadedFile] | None = None,
    ) -> GQLResponse:
        """
        Make a GraphQL query to the test client.

        :params query: GraphQL query string.
        :params operation_name: Name of the operation to execute.
        :params input_data: Set (and override) the "input" variable in the given variables.
        :params variables: Variables for the query.
        :params headers: Headers for the query. Keys should in all-caps, and be prepended with "HTTP_".
        :params files: Files to be uploaded with the query. Requires `graphene-file-upload`.
        """
        with capture_database_queries() as results:
            if files is not None:  # pragma: no cover
                from graphene_file_upload.django.testing import file_graphql_query

                response = file_graphql_query(
                    query=query,
                    op_name=operation_name,
                    input_data=input_data,
                    variables=variables,
                    headers=headers,
                    files=files,
                    client=self,
                )
            else:
                response = graphql_query(
                    query=query,
                    operation_name=operation_name,
                    input_data=input_data,
                    variables=variables,
                    headers=headers,
                    client=self,
                )
        return GQLResponse(response, results)

    def login_with_superuser(self) -> User:
        user = get_user_model().objects.create_superuser(username="superuser", email="superuser@django.com")
        self.force_login(user)
        return user

    def login_with_regular_user(self) -> User:
        user = get_user_model().objects.create_user(username="user", email="user@django.com")
        self.force_login(user)
        return user


@dataclass
class QueryData:
    queries: list[str]

    @property
    def log(self) -> str:
        message = "-" * 75
        message += f"\n>>> Queries ({len(self.queries)}):\n"
        for index, query in enumerate(self.queries):
            formatted_query = sqlparse.format(query, reindent=True)
            message += f"{index + 1}) ".ljust(75, "-") + f"\n{formatted_query}\n"
        message += "-" * 75
        return message


def _db_query_logger(  # noqa: PLR0913
    execute: Callable[..., Any],
    sql: str,
    params: tuple[Any, ...],
    many: bool,  # noqa: FBT001
    context: dict[str, Any],
    # Added with functools.partial()
    query_cache: list[str],
) -> Any:
    """
    A database query logger for capturing executed database queries.
    Used to check that query optimizations work as expected.

    Can also be used as a place to put debugger breakpoint for solving issues.
    """
    # Don't include transaction creation, as we aren't interested in them.
    if not sql.startswith("SAVEPOINT") and not sql.startswith("RELEASE SAVEPOINT"):
        try:
            query_cache.append(sql % params)
        except TypeError:  # pragma: no cover
            query_cache.append(sql)
    return execute(sql, params, many, context)


@contextmanager
def capture_database_queries() -> Generator[QueryData, None, None]:
    """Capture results of what database queries were executed. `DEBUG` needs to be set to True."""
    results = QueryData(queries=[])
    query_logger = partial(_db_query_logger, query_cache=results.queries)

    with db.connection.execute_wrapper(query_logger):
        yield results


TNamedTuple = TypeVar("TNamedTuple", bound=NamedTuple)


class ParametrizeArgs(TypedDict):
    argnames: list[str]
    argvalues: list[TNamedTuple]
    ids: list[str]


def parametrize_helper(__tests: dict[str, TNamedTuple], /) -> ParametrizeArgs:
    """Construct parametrize input while setting test IDs."""
    assert __tests, "I need some tests, please!"  # noqa: S101
    values = list(__tests.values())
    try:
        return ParametrizeArgs(
            argnames=list(values[0].__class__.__annotations__),
            argvalues=values,
            ids=list(__tests),
        )
    except AttributeError as error:  # pragma: no cover
        msg = "Improper configuration. Did you use a NamedTuple for TNamedTuple?"
        raise RuntimeError(msg) from error


def _dict_sorter(data: dict[str, Any]) -> tuple[str, str]:
    return str(data.keys()), str(data.values())


class ComparisonError(Exception):
    errors_by_code: dict[int, str] = {
        1: "type mismatch",
        2: "value mismatch",
        3: "dict keys mismatch",
        4: "array length mismatch",
        5: "array content mismatch",
        6: "array type mismatch",
    }

    def __init__(self, actual: Any, expected: Any, *, reason_code: int) -> None:
        self.actual = actual
        self.expected = expected
        self.actual_key = ""
        self.expected_key = ""
        self.reason = self.errors_by_code[reason_code]


def compare_unordered(actual: dict[str, Any] | list[Any], expected: dict[str, Any] | list[Any]) -> None:
    """
    Compare two json data structures without caring about the order in which items appear in lists.

    If the comparison fails, will raise a `pytest.fail` with an error message
    which will try to point out the place where the comparison failed.
    However, this will not always pick the best "nodes" to show, so
    full comparison data will also be shown in the error message for clarity.

    Disclaimer: Cannot compare json with heterogeneous lists.
    """
    msg: str = ""
    try:
        _compare_unordered(actual, expected)
    except ComparisonError as error:
        msg = cleandoc(
            f"""
            Actual and expected data don't match ({error.reason}):
                actual{error.actual_key} = {error.actual}
                expected{error.expected_key} = {error.expected}
            """
        )
        if error.actual_key or error.expected_key:
            msg = (
                msg
                + "\n"
                + cleandoc(
                    f"""
                    Full data:
                        actual = {actual}
                        expected = {expected}
                    """
                )
            )
    finally:
        if msg:
            pytest.fail(msg, pytrace=False)


def _compare_unordered(actual: Any, expected: Any) -> None:  # noqa: PLR0912, C901
    if type(actual) is not type(expected):
        raise ComparisonError(actual, expected, reason_code=1)

    if isinstance(actual, dict):
        if actual.keys() != expected.keys():
            raise ComparisonError(actual, expected, reason_code=3)

        for key in actual:
            try:
                _compare_unordered(actual[key], expected[key])
            except ComparisonError as error:  # noqa: PERF203
                error.actual_key = f"['{key}']{error.actual_key}"
                error.expected_key = f"['{key}']{error.expected_key}"
                raise

    elif isinstance(actual, list):
        if len(actual) != len(expected):
            raise ComparisonError(actual, expected, reason_code=4)

        actual_types = {type(item) for item in actual}
        expected_types = {type(item) for item in expected}

        if len(actual_types) != 1 or len(expected_types) != 1:
            raise ComparisonError(actual, expected, reason_code=6)

        actual_type = actual_types.pop()
        expected_type = expected_types.pop()
        if actual_type is not expected_type:
            raise ComparisonError(actual, expected, reason_code=5)

        if issubclass(actual_type, dict):
            for act, exp in zip(sorted(actual, key=_dict_sorter), sorted(expected, key=_dict_sorter)):
                try:
                    _compare_unordered(act, exp)
                except ComparisonError as error:  # noqa: PERF203
                    error.actual_key = f"[-]{error.actual_key}"
                    error.expected_key = f"[-]{error.expected_key}"
                    raise
        else:
            for act, exp in zip(sorted(actual), sorted(expected)):
                try:
                    _compare_unordered(act, exp)
                except ComparisonError as error:  # noqa: PERF203
                    error.actual_key = f"[-]{error.actual_key}"
                    error.expected_key = f"[-]{error.expected_key}"
                    raise

    elif actual != expected:
        raise ComparisonError(actual, expected, reason_code=2)



from __future__ import annotations

import sys
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Literal,
    Mapping,
    NamedTuple,
    ParamSpec,
    Protocol,
    Sequence,
    TypeAlias,
    TypedDict,
    TypeVar,
    Union,
)

# New in version 3.11
if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self

    class StrEnum(str, Enum):
        pass

else:  # pragma: no cover
    from enum import StrEnum
    from typing import Self

from django.core.handlers.wsgi import WSGIRequest
from graphql import GraphQLResolveInfo

if TYPE_CHECKING:
    import django_filters
    from django.contrib.auth.models import AnonymousUser, User
    from django.core.exceptions import ValidationError as DjangoValidationError
    from django.db.models import Field, Model, Q, QuerySet
    from django.forms import Form
    from graphql_relay import Edge, PageInfo
    from rest_framework.exceptions import ErrorDetail
    from rest_framework.exceptions import ValidationError as SerializerValidationError
    from rest_framework.serializers import ListSerializer

    from .bases import DjangoNode
    from .constants import Operation
    from .serializers import NestingModelSerializer


__all__ = [
    "Any",
    "AnyUser",
    "Callable",
    "FieldError",
    "FieldNameStr",
    "FilterOverride",
    "FilterSetMeta",
    "GQLInfo",
    "Generic",
    "GraphQLFilterInfo",
    "Iterable",
    "Literal",
    "Mapping",
    "NamedTuple",
    "NodeConnectionType",
    "ParamSpec",
    "PermCheck",
    "Protocol",
    "RelatedSerializer",
    "RelationType",
    "Self",
    "Sequence",
    "SerializerErrorData",
    "SerializerMeta",
    "StrEnum",
    "TypeAlias",
    "TypeVar",
    "TypedDict",
    "Union",
    "ValidationErrorType",
]


FieldNameStr: TypeAlias = str
FieldLookupStr: TypeAlias = str
FilterAliasStr: TypeAlias = str
LookupNameStr: TypeAlias = str
MethodNameStr: TypeAlias = str
AnyUser: TypeAlias = Union["User", "AnonymousUser"]
PermCheck: TypeAlias = Callable[[AnyUser], bool] | Callable[[AnyUser, "Model"], bool]
RelationType: TypeAlias = Literal["one_to_one", "many_to_one", "one_to_many", "many_to_many"]
RelatedSerializer: TypeAlias = Union["NestingModelSerializer", "ListSerializer"]
Fields: TypeAlias = Sequence[FieldNameStr] | Literal["__all__"]
FieldAliasToLookup: TypeAlias = dict[FilterAliasStr, FieldLookupStr]
FilterFields: TypeAlias = Sequence[FieldLookupStr | tuple[FieldLookupStr, FilterAliasStr]] | Literal["__all__"]
ValidationErrorType: TypeAlias = Union["DjangoValidationError", "SerializerValidationError"]
SerializerErrorData: TypeAlias = dict[str, list[Union["ErrorDetail", str]]]

T = TypeVar("T")


class NodeConnectionType(Protocol[T]):
    page_info: PageInfo
    edged: list[Edge]
    iterable: QuerySet[T]
    length: int


class OrderingFunc(Protocol):
    def __call__(self, qs: QuerySet, *, desc: bool) -> QuerySet:
        """Custom ordering function."""


class FilterOverride(TypedDict):
    filter_class: type[django_filters.Filter]
    extra: Callable[[Field], dict[str, Any]]


class UserHintedWSGIRequest(WSGIRequest):
    user: AnyUser


class GQLInfo(GraphQLResolveInfo):
    context = UserHintedWSGIRequest


class SerializerMeta:
    model: type[Model]
    fields: Fields
    read_only_fields: Sequence[FieldNameStr]
    exclude: Sequence[FieldNameStr]
    depth: int
    extra_kwargs: Mapping[FieldNameStr, Mapping[str, Any]]
    node: DjangoNode


class FilterSetMeta:
    model: type[Model] | None
    fields: Fields | Mapping[FieldNameStr, Sequence[LookupNameStr]] | None
    exclude: Sequence[FieldNameStr] | None
    filter_overrides: Mapping[Field, FilterOverride]
    form: type[Form]
    combination_methods: Sequence[MethodNameStr]
    order_by: Sequence[FieldNameStr | tuple[FieldLookupStr, FilterAliasStr]]


@dataclass
class UserDefinedFilterInput:
    operation: Operation
    field: Enum | str | None = None
    value: Any = None
    operations: list[UserDefinedFilterInput] | None = None


@dataclass
class UserDefinedFilterResult:
    filters: Q
    annotations: dict[str, Any]
    ordering: list[str]


class FieldError(TypedDict):
    field: str
    message: str
    code: str


class GraphQLFilterInfo(TypedDict, total=False):
    name: str
    filters: dict[str, Any]
    children: dict[str, GraphQLFilterInfo]



from __future__ import annotations

import pytest

__all__ = [
    "graphql",
    "query_counter",
]


@pytest.fixture()
def graphql():  # noqa: ANN201
    from graphene_django_extensions.testing import GraphQLClient

    return GraphQLClient()


@pytest.fixture()
def query_counter():  # pragma: no cover  # noqa: ANN201
    from graphene_django_extensions.testing.client import capture_database_queries

    return capture_database_queries



from __future__ import annotations

from typing import TYPE_CHECKING

from graphene.types.mutation import MutationOptions
from graphene_django.types import DjangoObjectTypeOptions

if TYPE_CHECKING:
    from django.db import models
    from django.forms import Form, ModelForm
    from rest_framework.serializers import ModelSerializer, Serializer

    from graphene_django_extensions.permissions import BasePermission

    from .bases import DjangoMutation, DjangoNode
    from .typing import FieldNameStr, Literal, Sequence


__all__ = [
    "DjangoNodeOptions",
    "DjangoMutationOptions",
]


class DjangoNodeOptions(DjangoObjectTypeOptions):
    def __init__(
        self,
        class_type: type[DjangoNode],
        max_complexity: int,
        permission_classes: Sequence[type[BasePermission]] = (),
    ) -> None:
        self.max_complexity = max_complexity
        self.permission_classes = permission_classes
        super().__init__(class_type)


class DjangoMutationOptions(MutationOptions):
    def __init__(  # noqa: PLR0913
        self,
        class_type: type[DjangoMutation],
        lookup_field: FieldNameStr | None,
        model_class: type[models.Model] | None,
        serializer_class: type[ModelSerializer | Serializer] | None,
        output_serializer_class: type[ModelSerializer | Serializer] | None,
        form_class: type[ModelForm | Form] | None,
        output_form_class: type[ModelForm | Form] | None,
        permission_classes: Sequence[type[BasePermission]],
        model_operation: Literal["create", "update", "delete", "custom"],
    ) -> None:
        self.lookup_field = lookup_field
        self.model_class = model_class
        self.model_operation = model_operation
        self.serializer_class = serializer_class
        self.output_serializer_class = output_serializer_class
        self.form_class = form_class
        self.output_form_class = output_form_class
        self.permission_classes = permission_classes
        super().__init__(class_type)



from __future__ import annotations

from typing import TYPE_CHECKING

import graphene

if TYPE_CHECKING:
    from .typing import Any, GQLInfo, NodeConnectionType

__all__ = [
    "Connection",
]


class Connection(graphene.Connection):
    """Connection field that adds the `totalCount` field to the connection."""

    class Meta:
        abstract = True

    total_count = graphene.Int()

    def resolve_total_count(root: NodeConnectionType, info: GQLInfo, **kwargs: Any) -> int:  # noqa: N805
        return root.length



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

if TYPE_CHECKING:
    from django.forms import Field

    from ..typing import Any, Sequence

__all__ = [
    "StrChoiceField",
    "IntChoiceField",
]


class StrChoiceField(models.CharField):  # pragma: no cover
    """CharField for TextChoices that automatically sets 'max_length' to the length of the longest choice."""

    def __init__(self, enum: type[models.Choices], **kwargs: Any) -> None:
        self.enum = enum
        kwargs["max_length"] = max(len(val) for val, _ in enum.choices)
        kwargs["choices"] = enum.choices
        super().__init__(**kwargs)

    def deconstruct(self) -> tuple[str, str, Sequence[Any], dict[str, Any]]:
        name, path, args, kwargs = super().deconstruct()
        kwargs["enum"] = self.enum
        return name, path, args, kwargs


class IntChoiceField(models.IntegerField):  # pragma: no cover
    """
    IntegerField for IntegerChoices that sets 'min_value' and 'max_value' validators
    instead of regular choice validation. Useful when you don't want integer choices to be
    converted to strings in graphql endpoints (ref: `graphene_django.converter.convert_choice_name`).
    Pair with `IntChoiceField` to validate choices in serializers.
    """

    def __init__(self, enum: type[models.Choices], **kwargs: Any) -> None:
        self.enum = enum

        min_value = int(min(val for val, _ in enum.choices))
        max_value = int(max(val for val, _ in enum.choices))
        msg = f"Value must be between {min_value} and {max_value}."
        # These validators are passed to the serializer field as its 'min_value' and 'max_value'.
        kwargs["validators"] = [
            MinValueValidator(limit_value=min_value, message=msg),
            MaxValueValidator(limit_value=max_value, message=msg),
        ]
        super().__init__(**kwargs)

    def deconstruct(self) -> tuple[str, str, Sequence[Any], dict[str, Any]]:
        name, path, args, kwargs = super().deconstruct()
        kwargs["enum"] = self.enum
        return name, path, args, kwargs

    def validate(self, value: int | None, model_instance: Any) -> None:
        original_choices = self.choices
        try:
            self.choices = self.enum.choices
            super().validate(value, model_instance)
        finally:
            self.choices = original_choices

    def formfield(self, **kwargs: Any) -> Field:
        original_choices = self.choices
        try:
            self.choices = self.enum.choices
            choices = super().formfield(**kwargs)
        finally:
            self.choices = original_choices

        return choices



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

import graphene
from django import forms  # noqa: TCH002
from django.db import models  # noqa: TCH002
from graphene.types.enum import Enum  # noqa: TCH002
from graphene_django.converter import (
    convert_choices_to_named_enum_with_descriptions,
    convert_django_field,
    get_django_field_description,
)
from graphene_django.forms.converter import convert_form_field, get_form_field_description
from graphene_django.registry import Registry  # noqa: TCH002
from graphene_django.rest_framework.serializer_converter import get_graphene_type_from_serializer_field
from rest_framework.fields import ChoiceField  # noqa: TCH002
from rest_framework.serializers import ListSerializer, ModelSerializer

from .fields import (
    Duration,
    DurationField,
    EnumChoiceField,
    EnumMultipleChoiceField,
    IntChoiceField,
    IntMultipleChoiceField,
    OrderByField,
    OrderingChoices,
    Time,
    UserDefinedFilterField,
    UserDefinedFilterInputType,
)
from .model_operations import get_model_lookup_field

if TYPE_CHECKING:
    from django.forms import Field, Form, ModelForm

    from .typing import FieldNameStr, SerializerMeta


__all__ = [
    "convert_form_fields_to_not_required",
    "convert_serializer_fields_to_not_required",
]


def convert_serializer_fields_to_not_required(
    serializer_class: type[ModelSerializer],
    lookup_field: FieldNameStr | None,
    *,
    top_level: bool = True,
) -> type[ModelSerializer | ListSerializer]:
    """
    When updating, usually the wanted behaviour is that the user can only update the
    fields that are actually updated. Therefore, this function can be used to create a new
    serializer class, on which has all the appropriate fields set to `required=False` (except
    the top-level `lookup_field`, which is required to select the row to be updated).
    This function is recursive, so that nested serializers, and their fields are also converted
    to `required=False`.

    :param serializer_class: The serializer class to convert.
    :param lookup_field: The lookup field to be used for the update operation.
    :param top_level: Whether this is the top-level serializer.
    """
    # We need to create a new serializer and rename it since
    # `graphene_django.rest_framework.serializer_converter.convert_serializer_to_input_type`.
    # caches its results by the serializer class name, and thus the input type created for
    # the CREATE operation would be used if created first.
    class_name = f"Update{serializer_class.__name__}"
    new_class: type[ModelSerializer] = type(class_name, (serializer_class,), {})  # type: ignore[assignment]

    # Create a new Meta and deepcopy `extra_kwargs` to avoid changing the original serializer,
    # which might be used for other operations.
    new_class.Meta: SerializerMeta = type("Meta", (serializer_class.Meta,), {})  # type: ignore[assignment]
    new_class.Meta.extra_kwargs = deepcopy(getattr(serializer_class.Meta, "extra_kwargs", {}))

    lookup_field = get_model_lookup_field(new_class.Meta.model, lookup_field)

    for field_name in new_class.Meta.fields:
        _set_extra_kwargs(field_name, new_class.Meta, lookup_field, top_level=top_level)

    # Handle nested serializers
    for field_name, field in new_class._declared_fields.items():
        if isinstance(field, ModelSerializer):
            new_field = convert_serializer_fields_to_not_required(field.__class__, None, top_level=False)
            new_kwargs = field._kwargs | {"required": False}
            new_class._declared_fields[field_name] = new_field(*field._args, **new_kwargs)

        elif isinstance(field, ListSerializer) and isinstance(field.child, ModelSerializer):
            new_child = convert_serializer_fields_to_not_required(field.child.__class__, None, top_level=False)
            new_kwargs = field.child._kwargs | {"required": False, "many": True}
            new_class._declared_fields[field_name] = new_child(*field.child._args, **new_kwargs)

    return new_class


def _set_extra_kwargs(
    field_name: FieldNameStr,
    meta: SerializerMeta,
    lookup_field: FieldNameStr,
    *,
    top_level: bool,
) -> None:
    """Set `meta.extra_kwargs` settings for the given `field_name`."""
    # If the `field_name` is for a model property, should not set anything.
    attr = getattr(meta.model, field_name, None)
    if isinstance(attr, property) and field_name != lookup_field:
        return

    # If the `field_name` is the given `lookup field` while `top_level=True`, should set `required=True`,
    # otherwise should set `required=False`, unless the value has been explicitly set already.
    required = top_level and field_name == lookup_field
    field_options = meta.extra_kwargs.setdefault(field_name, {})
    field_options.setdefault("required", required)

    # Lookup field should be additionally marked as writeable so that
    # serializer doesn't remove it during validation.
    if field_name == lookup_field:
        meta.extra_kwargs.setdefault(field_name, {})["read_only"] = False


def convert_form_fields_to_not_required(form_class: type[ModelForm | Form]) -> type[ModelForm | Form]:
    form_fields: dict[str, Field] = {}
    for field_name, field in form_class.base_fields.items():
        new_field = deepcopy(field)
        new_field.required = False
        form_fields[field_name] = new_field

    return type(form_class.__name__, (form_class,), form_fields)  # type: ignore[return-value]


@convert_django_field.register
def convert_time_to_string(field: models.TimeField, registry: Registry | None = None) -> Time:
    return Time(description=get_django_field_description(field), required=not field.null)


@convert_form_field.register
def convert_form_field_to_time(field: forms.TimeField) -> Time:
    return Time(description=get_form_field_description(field), required=field.required)


@convert_django_field.register
def convert_duration_to_int(field: models.DurationField, registry: Registry | None = None) -> Duration:
    return Duration(description=get_django_field_description(field), required=not field.null)


@convert_form_field.register
def convert_form_field_to_duration(field: forms.DurationField) -> Duration:
    return Duration(description=get_form_field_description(field), required=field.required)


@convert_form_field.register
def convert_form_field_to_int(field: IntChoiceField) -> graphene.Int:
    return graphene.Int(description=get_form_field_description(field), required=field.required)


@convert_form_field.register
def convert_form_field_to_list_of_int(field: IntMultipleChoiceField) -> graphene.List:
    return graphene.List(graphene.Int, description=get_form_field_description(field), required=field.required)


@convert_form_field.register
def convert_form_field_to_enum(field: EnumChoiceField) -> graphene.Enum:
    return graphene.Enum.from_enum(field.enum)(
        description=get_form_field_description(field),
        required=field.required,
    )


@convert_form_field.register
def convert_form_field_to_enum_list(field: EnumMultipleChoiceField) -> graphene.List:
    return graphene.List(
        graphene.Enum.from_enum(field.enum),
        description=get_form_field_description(field),
        required=field.required,
    )


@convert_form_field.register
def convert_user_defined_filter(field: UserDefinedFilterField) -> UserDefinedFilterInputType:
    return UserDefinedFilterInputType.create(
        model=field.model,
        fields_map=field.fields_map,
    )(
        description=get_form_field_description(field),
        required=field.required,
    )


@convert_form_field.register
def convert_ordering_field(field: OrderByField) -> graphene.List:
    return graphene.List(
        OrderingChoices.create(model=field.model, fields_map=field.fields_map),
        description=get_form_field_description(field),
        required=field.required,
    )


@get_graphene_type_from_serializer_field.register
def convert_serializer_field_to_duration(field: DurationField) -> type[Duration]:
    return Duration


@get_graphene_type_from_serializer_field.register
def convert_serializer_field_to_enum(field: ChoiceField) -> Enum:
    # `EnumFriendlyChoiceField` can have a reference to the enum class for consistent naming
    # of the same enum in different places (nodes, mutations, filters, etc.)
    if hasattr(field, "enum") and field.enum is not None:  # pragma: no cover
        name = field.enum.__name__
    else:
        name = field.field_name or field.source or "Choices"
        name = "".join(s.capitalize() for s in name.split("_"))
    return convert_choices_to_named_enum_with_descriptions(name, field.choices)



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

from django.db import IntegrityError, models, transaction
from graphene_django.types import ALL_FIELDS
from rest_framework.exceptions import ValidationError
from rest_framework.serializers import ListSerializer, ModelSerializer

from .errors import get_constraint_message
from .fields import DurationField, EnumFriendlyChoiceField, IntegerPrimaryKeyField
from .model_operations import RelatedFieldInfo, get_object_or_404, get_related_field_info
from .typing import ParamSpec, SerializerMeta, TypeVar
from .utils import add_translatable_fields

if TYPE_CHECKING:
    from django.db.models import Model
    from rest_framework.fields import Field

    from .typing import Any, AnyUser, Callable, RelatedSerializer, Self


__all__ = [
    "NestingModelSerializer",
]


T = TypeVar("T")
P = ParamSpec("P")


def _related_pre_and_post_save(func: Callable[P, T]) -> Callable[P, T]:
    """Handle related models before and after creating or updating the main model."""

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> models.Model:
        self: NestingModelSerializer = args[0]
        validated_data = next((arg for arg in args if isinstance(arg, dict)), kwargs.get("validated_data"))
        if validated_data is None:  # pragma: no cover
            msg = "'validated_data' not found in args or kwargs"
            raise ValueError(msg)

        try:
            with transaction.atomic():
                related_serializers = self._pre_save(validated_data)
                instance = func(*args, **kwargs)
                if related_serializers:
                    self._post_save(instance, related_serializers)
        except IntegrityError as error:
            msg = get_constraint_message(error.args[0])
            raise ValidationError(msg) from error

        return instance

    return wrapper


class NestingModelSerializer(ModelSerializer):
    """
    ModelSerializer that contains logic for updating and creating related models
    when they are included as nested serializer fields:

    ```
    from graphene_django_extensions import NestingModelSerializer

    class SubSerializer(NestingModelSerializer):
        class Meta:
            model = Sub
            fields = ["pk", "field"]

    class MainSerializer(NestingModelSerializer):
        sub_entry = SubSerializer()

        class Meta:
            model = Main
            fields = ["pk", "sub_entry"]
    ```

    When using the above serializers with the following data:

    ```
    {
        "pk": 1,
        "sub_entities": {
            "field": "value"
        }
    }
    ```

    This will create the Main entity and the Sub entity,
    and set the Sub entity's 'main' field to the Main entity.

    If instead this is used:

    ```
    {
        "pk": 1,
        "sub_entities": {
            "pk": 2
        }
    }
    ```

    This will create the Main entity and link an existing Sub entity with pk=2 to it.
    If the Sub entity does not exist, a 404 error will be raised. If the sub entity
    is already linked to another Main entity, this will be a no-op.

    If the 'pk' field is included with some other fields, the Sub entity will also be updated:

    ```
    {
        "pk": 1,
        "sub_entities": {
            "pk": 2,
            "field": "value"
        }
    }
    ```

    For "to_many" relations, the serializer field must be a ListSerializer:

    ```
    class MainSerializer(NestingModelSerializer):
        sub_entry = SubSerializer(many=True)
    ```

    Same logic applies for "to_many" relations, but if the relation is a 'one_to_many' relation,
    and the relation is updated. Any existing related entities that were not included in the request
    will be deleted (e.g., `pk=1` could have been deleted here):

    ```
    {
        "pk": 1,
        "sub_entities": [
            {"pk": 2},
            {"field": "value"}
        ]
    }
    ```
    """

    instance: Model  # Use this to hint the instance model type in subclasses
    serializer_related_field = IntegerPrimaryKeyField  # Related fields defined in Meta are integers by default
    serializer_choice_field = EnumFriendlyChoiceField  # Converts enums to string correctly
    serializer_field_mapping = ModelSerializer.serializer_field_mapping | {
        models.DurationField: DurationField,
    }

    class Meta(SerializerMeta):
        pass

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        if cls.Meta.fields != ALL_FIELDS:
            cls.Meta.fields = add_translatable_fields(cls.Meta.model, cls.Meta.fields)
        return super().__new__(cls, *args, **kwargs)

    def get_or_default(self, field: str, attrs: dict[str, Any]) -> Any:
        """
        Get field value from attrs, or if not found, use a default value from
        1) the serializer model instance, or 2) the model field.
        """
        default = self.Meta.model._meta.get_field(field).default
        default = getattr(self.instance, field, default)
        return attrs.get(field, default)

    @property
    def request_user(self) -> AnyUser:
        return self.context["request"].user

    def get_update_or_create(self, data: dict[str, Any] | None) -> Model | None:
        if data is None:  # pragma: no cover
            return None

        pk = data.pop("pk", None)
        if pk is not None:
            instance = get_object_or_404(self.Meta.model, pk=pk)
            if not data:
                return instance
            return self.update(instance, data)

        return self.create(data)

    @_related_pre_and_post_save
    def create(self, validated_data: dict[str, Any]) -> Model:
        """Create a new instance of the model, while also handling related models."""
        return self.Meta.model._default_manager.create(**validated_data)

    @_related_pre_and_post_save
    def update(self, instance: Model, validated_data: dict[str, Any]) -> Model:
        """Update an existing instance of the model, while also handling related models."""
        for attr, value in validated_data.items():
            setattr(instance, attr, value)
        instance.save()
        return instance

    def _pre_save(self, validated_data: dict[str, Any]) -> dict[str, RelatedSerializer]:
        """
        Prepare related models defined using BaseModelSerializers.
        Forward 'one-to-one' and 'many-to-one' related entities will be fetched, updated, or created.
        Other related entities will be saved to be handled after the main model is saved using '_handle_to_many'.
        """
        related_serializers: dict[str, RelatedSerializer] = {}
        related_info = get_related_field_info(self.Meta.model)

        for name in list(validated_data):  # Copy keys so that we can pop from the original dict in the loop
            field: Field | None = self.fields.get(name, None)
            if field is None:
                continue

            related_field_info = related_info.get(name, None)
            if related_field_info is None:
                continue

            # Handle relation types:
            #  - Forward one-to-one
            #  - Forward many-to-one
            #  - Reverse one-to-one
            if isinstance(field, NestingModelSerializer):
                if related_field_info.reverse and related_field_info.one_to_one:
                    field.initial_data = validated_data.pop(name, None)
                    if field.initial_data is None:  # pragma: no cover
                        continue

                    field.related_field_info = related_field_info
                    related_serializers[name] = field
                    continue

                rel_data = validated_data.pop(name, None)
                validated_data[name] = field.get_update_or_create(rel_data)

            # Handle relation types:
            #  - Reverse one-to-many
            #  - Reverse many-to-many
            #  - Forward many-to-many
            elif isinstance(field, ListSerializer) and isinstance(field.child, NestingModelSerializer):
                field.initial_data = validated_data.pop(name, None)
                if field.initial_data is None:  # pragma: no cover
                    continue

                field.related_field_info = related_field_info
                related_serializers[name] = field

        return related_serializers

    def _post_save(self, instance: Model, related_serializers: dict[str, RelatedSerializer]) -> None:
        """
        Handle creating or updating related models after the main model.
        Delete any existing 'one_to_many' entities that were untouched in this request.
        Add any new 'many_to_many' entities that were not previously linked to the main model.
        """
        for field_name, serializer in related_serializers.items():
            rel_info: RelatedFieldInfo | None = getattr(serializer, "related_field_info", None)
            if rel_info is None:  # pragma: no cover
                continue

            # Handle reverse one-to-one
            if isinstance(serializer, NestingModelSerializer):
                serializer.initial_data[rel_info.name] = instance
                serializer.get_update_or_create(serializer.initial_data)
                continue

            instances: list[Model] = []
            pks: list[Any] = []
            child_serializer: NestingModelSerializer = serializer.child

            for item in serializer.initial_data:
                if rel_info.one_to_many:
                    item[rel_info.name] = instance

                nested_instance = child_serializer.get_update_or_create(item)
                if nested_instance is None:  # pragma: no cover
                    continue

                if rel_info.one_to_many:
                    pks.append(nested_instance.pk)
                if rel_info.many_to_many:
                    instances.append(nested_instance)

            if rel_info.one_to_many:
                # Delete related objects that were not created or modified.
                child_serializer.Meta.model.objects.filter(**{rel_info.name: instance}).exclude(pk__in=pks).delete()

            if rel_info.many_to_many:
                # Add related objects that were not previously linked to the main model.
                getattr(instance, field_name).set(instances)

    def get_fields(self) -> dict[str, Field]:
        fields = super().get_fields()
        # Add the model field enum to `EnumFriendlyChoiceField`
        # if the enum was not explicitly defined in the serializer field.
        # (e.g. field created automatically based on `serializer_choice_field`)
        # This is used for mutation input enum naming.
        for name, field in fields.items():
            if hasattr(field, "enum") and field.enum is None:
                model_field = self.Meta.model._meta.get_field(name)
                field.enum = getattr(model_field, "enum", None)
        return fields



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms
from graphene.utils.str_converters import to_camel_case

if TYPE_CHECKING:
    from django.db.models import Choices, Model

    from ..typing import Any, FieldAliasToLookup, FieldNameStr

__all__ = [
    "IntChoiceField",
    "IntMultipleChoiceField",
    "EnumChoiceField",
    "EnumMultipleChoiceField",
    "UserDefinedFilterField",
    "OrderByField",
]


class IntChoiceFieldMixin:
    def __init__(self: forms.TypedChoiceField | forms.TypedMultipleChoiceField, **kwargs: Any) -> None:
        kwargs["coerce"] = int
        kwargs["empty_value"] = None
        super().__init__(**kwargs)

    def valid_value(self: forms.TypedChoiceField | forms.TypedMultipleChoiceField, value: Any) -> bool:
        if self.choices:  # pragma: no cover
            parent: forms.TypedChoiceField = super()  # type: ignore[assignment]
            return parent.valid_value(value)
        try:
            self.coerce(value)
        except (ValueError, TypeError):  # pragma: no cover
            return False
        return True


class EnumChoiceFieldMixin:
    def __init__(self, enum: type[Choices], **kwargs: Any) -> None:
        self.enum = enum
        kwargs["choices"] = enum.choices
        super().__init__(**kwargs)


class IntChoiceField(IntChoiceFieldMixin, forms.TypedChoiceField):
    """Allow plain integers as choices in GraphQL filters. Supports a single choice."""


class IntMultipleChoiceField(IntChoiceFieldMixin, forms.TypedMultipleChoiceField):
    """Same as `IntChoiceField` above but supports multiple choices."""


class EnumChoiceField(EnumChoiceFieldMixin, forms.ChoiceField):
    """
    Custom field for handling enums better in GraphQL filters. Supports a single choice.
    See `EnumChoiceFilter` for motivation.
    """


class EnumMultipleChoiceField(EnumChoiceFieldMixin, forms.MultipleChoiceField):
    """Same as `EnumChoiceField` but supports multiple choices."""


class UserDefinedFilterField(forms.Field):
    """Used together with `UserDefinedFilter`."""

    def __init__(self, model: type[Model], fields: FieldAliasToLookup, **kwargs: Any) -> None:
        self.model = model
        self.fields_map = fields
        super().__init__(**kwargs)


class OrderByField(forms.Field):
    """Used together with `CustomOrderingFilter`."""

    def __init__(self, model: type[Model], choices: list[tuple[FieldNameStr, str]], **kwargs: Any) -> None:
        self.model = model
        self.fields_map: dict[str, str] = {
            to_camel_case(name[1:]) + "Desc" if name[0] == "-" else to_camel_case(name) + "Asc": name
            for name, _ in choices
        }
        kwargs.pop("null_label", None)
        super().__init__(**kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

import django_filters
from django import forms
from django.db import models
from django.db.models import Model, Q, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django_filters.constants import ALL_FIELDS, EMPTY_VALUES
from django_filters.filterset import FILTER_FOR_DBFIELD_DEFAULTS
from graphene.utils.str_converters import to_camel_case
from query_optimizer.filter import FilterSet

from .fields import (
    EnumChoiceField,
    EnumMultipleChoiceField,
    IntChoiceField,
    IntMultipleChoiceField,
    OrderByField,
    UserDefinedFilterField,
)
from .settings import gdx_settings
from .typing import (
    FieldAliasToLookup,
    FilterFields,
    FilterSetMeta,
    UserDefinedFilterInput,
    UserDefinedFilterResult,
)

if TYPE_CHECKING:
    from .typing import Any, OrderingFunc

__all__ = [
    "CustomOrderingFilter",
    "EnumChoiceFilter",
    "EnumMultipleChoiceFilter",
    "IntChoiceFilter",
    "IntMultipleChoiceFilter",
    "ModelFilterSet",
    "UserDefinedFilter",
]


class EnumChoiceFilterMixin:
    def __init__(self, enum: type[models.Choices], *args: Any, **kwargs: Any) -> None:
        kwargs["enum"] = enum
        kwargs["choices"] = enum.choices
        super().__init__(*args, **kwargs)


class IntChoiceFilter(django_filters.TypedChoiceFilter):
    """
    Allow plain integers as choices in GraphQL filters.
    Normally, integer enums are converted to string enums in GraphQL by prefixing
    them with `A_`, but this filter allows using plain integers.
    """

    field_class = IntChoiceField


class IntMultipleChoiceFilter(django_filters.TypedMultipleChoiceFilter):
    """Same as above but supports multiple choices."""

    field_class = IntMultipleChoiceField


class EnumChoiceFilter(EnumChoiceFilterMixin, django_filters.TypedChoiceFilter):
    """
    Custom field for handling enums better in GraphQL filters.
    Supports a single choice.

    Using `django_filters.ChoiceFilter` causes the enum choices to be converted to strings in GraphQL filters.
    This class uses GraphQL enums instead, which gives better autocomplete results.
    """

    field_class = EnumChoiceField


class EnumMultipleChoiceFilter(EnumChoiceFilterMixin, django_filters.TypedMultipleChoiceFilter):
    """Same as above but supports multiple choices."""

    field_class = EnumMultipleChoiceField


class UserDefinedFilter(django_filters.Filter):
    """Allows for user defined filter operations."""

    field_class = UserDefinedFilterField

    def __init__(self, model: type[Model], fields: FilterFields = ALL_FIELDS, **kwargs: Any) -> None:
        kwargs["model"] = model
        kwargs["fields"] = self._normalize_fields(model, fields)
        super().__init__(**kwargs)

    def filter(self, qs: QuerySet, data: UserDefinedFilterInput) -> QuerySet:
        if data in EMPTY_VALUES:
            return qs

        result = self.build_user_defined_filters(data)

        if result.annotations:  # pragma: no cover
            qs = qs.annotate(**result.annotations)
        if result.ordering:  # pragma: no cover
            qs = qs.order_by(*qs.query.order_by, *result.ordering)

        return qs.filter(result.filters)

    def build_user_defined_filters(self, data: UserDefinedFilterInput) -> UserDefinedFilterResult:
        filters: Q = Q()
        ann: dict[str, Any] = {}
        ordering: list[str] = []

        if data.operation.value in ("AND", "OR", "XOR", "NOT"):
            if data.operations is None:
                msg = "Logical filter operation requires 'operations' to be set."
                raise ValueError(msg)

            if data.operation.value == "NOT" and len(data.operations) != 1:
                msg = "Logical filter operation 'NOT' requires exactly one operation."
                raise ValueError(msg)

            for operation in data.operations:
                result = self.build_user_defined_filters(operation)
                if result.annotations:  # pragma: no cover
                    ann.update(result.annotations)
                if result.ordering:  # pragma: no cover
                    ordering.extend(result.ordering)

                if data.operation.value == "AND":
                    filters &= result.filters
                elif data.operation.value == "OR":
                    filters |= result.filters
                elif data.operation.value == "NOT":
                    filters = ~result.filters
                elif data.operation.value == "XOR":
                    filters ^= result.filters

        else:
            if data.field is None:
                msg = "Comparison filter operation requires 'field' to be set."
                raise ValueError(msg)

            alias: str = getattr(data.field, "name", data.field)
            field: str = self.extra["fields"][alias]
            inputs: dict[str, Any] = {f"{field}{LOOKUP_SEP}{data.operation.value.lower()}": data.value}
            filters = Q(**inputs)

        return UserDefinedFilterResult(filters=filters, annotations=ann, ordering=ordering)

    @staticmethod
    def _normalize_fields(model: type[Model], fields: FilterFields) -> FieldAliasToLookup:
        if fields == ALL_FIELDS:  # pragma: no cover
            return {to_camel_case(field.name): field.name for field in model._meta.get_fields()}

        normalized_fields: FieldAliasToLookup = {}
        for field in fields:
            if isinstance(field, tuple):
                normalized_fields[to_camel_case(field[1])] = field[0]
            else:
                normalized_fields[to_camel_case(field)] = field
        return normalized_fields


class CustomOrderingFilter(django_filters.OrderingFilter):
    """
    Ordering filter for handling custom orderings by defining `order_by_{name}` functions
    on its subclasses or filtersets it is defined on.
    """

    base_field_class = OrderByField
    field_class = forms.Field

    def filter(self, qs: models.QuerySet, value: list[str]) -> models.QuerySet:
        if value in EMPTY_VALUES:
            return qs

        ordering: list[str] = list(qs.query.order_by)
        for param in value:
            if param in EMPTY_VALUES:  # pragma: no cover
                continue

            func_name = f"order_by_{param.removeprefix('-')}"

            # Try to find an `ordering_func` on the `OrderingFilter` class or its `FilterSet` class.
            ordering_func: OrderingFunc | None = getattr(self, func_name, None)
            if ordering_func is None and hasattr(self, "parent"):
                ordering_func = getattr(self.parent, func_name, None)

            # If no `ordering_func` was found, just order by the given field name.
            if ordering_func is None or not callable(ordering_func):
                ordering.append(self.get_ordering_value(param))
                continue

            qs = ordering_func(qs, desc=param.startswith("-"))
            # Save the `order_by` value since the `qs.order_by(*ordering)`
            # will clear all ordering when called.
            ordering.extend(qs.query.order_by)

        return qs.order_by(*ordering)


class ModelFilterSet(FilterSet):
    """
    Custom FilterSet class for optimizing the filtering of GraphQL queries.
    Adds the following features to all types that inherit it:

    - Adds a default ordering filter by primary key if none is defined.

    - Changes the default filters for all relationships to not make a database
      query to check if a filtered rows exists.

    ---

    The following options can be set in the `Meta`-class.

    `model: type[models.Model]`

    - Required. Model class for the model the filterset is for.

    `fields: Sequence[FieldNameStr] | Mapping[FieldNameStr, Sequence[LookupNameStr]] | Literal["__all__"]`

    - Required if no `exclude`. Fields to include in the filterset. Can be a list of field names
      (lookup name will be `exact`), a mapping of field names to a list of lookup names, or the
      special value `__all__` to include all fields.

    `exclude: Sequence[FieldNameStr]`

    - Required if no `fields`. Fields to exclude from the filterset.

    `filter_overrides: dict[Field, FilterOverride]`

    - Optional. Overrides for the default filters for specific fields.

    `form: type[Form]`

    - Optional. Form class to use for the filterset. Defaults to `django_filters.Form`.

    `order_by: Sequence[FieldLookupStr | tuple[FieldLookupStr, FilterAliasStr]]`

    - Optional. Ordering filters to add to the filterset. Can also add non-field orderings,
      or customize field orderings by adding a `order_by_{field_name}` method to the
      `ModelFilterSet` subclass.

    `combination_methods: Sequence[MethodNameStr]`

    - Optional. Allows combining method filters so that they will use the same filter function.
      The combination method will always run, and its value will be a mapping of the
      field names of the combined filters to their values.
    """

    _meta: FilterSetMeta
    declared_filters: dict[str, django_filters.Filter]

    FILTER_DEFAULTS = FILTER_FOR_DBFIELD_DEFAULTS

    # Change the default filters for all relationships to not make
    # a database query to check if a filtered rows exists.
    FILTER_DEFAULTS[models.ForeignKey] = {"filter_class": IntChoiceFilter}
    FILTER_DEFAULTS[models.OneToOneField] = {"filter_class": IntChoiceFilter}
    FILTER_DEFAULTS[models.ManyToManyField] = {"filter_class": IntMultipleChoiceFilter}
    FILTER_DEFAULTS[models.OneToOneRel] = {"filter_class": IntChoiceFilter}
    FILTER_DEFAULTS[models.ManyToOneRel] = {"filter_class": IntMultipleChoiceFilter}
    FILTER_DEFAULTS[models.ManyToManyRel] = {"filter_class": IntMultipleChoiceFilter}

    class Meta(FilterSetMeta):
        pass

    @classmethod
    def get_filters(cls) -> dict[str, django_filters.Filter]:
        if cls._meta.model is not None:
            # Add a default ordering filter if none is defined, or extend an existing one.
            ordering_fields: list[str | tuple[str, str]] = getattr(cls.Meta, "order_by", ["pk"])
            ordering_filter = cls.declared_filters.get(gdx_settings.ORDERING_FILTER_NAME)

            if isinstance(ordering_filter, CustomOrderingFilter):
                # Param map should be flipped so that OrderingFilter initializes correctly.
                fields_map: dict[str, str] = {v: k for k, v in ordering_filter.param_map.items()}
                for field in ordering_fields:
                    if isinstance(field, tuple):
                        fields_map.setdefault(field[0], field[1])
                    else:
                        fields_map.setdefault(field, field)

                cls.declared_filters[gdx_settings.ORDERING_FILTER_NAME] = CustomOrderingFilter(
                    model=cls._meta.model,
                    fields=fields_map,
                )

            elif ordering_filter is None:
                cls.declared_filters[gdx_settings.ORDERING_FILTER_NAME] = CustomOrderingFilter(
                    model=cls._meta.model,
                    fields=ordering_fields,
                )

        if cls._meta.fields is None and cls.declared_filters:
            cls._meta.fields = []

        return super().get_filters()

    def filter_queryset(self, queryset: models.QuerySet) -> models.QuerySet:
        combination_methods: list[str] = getattr(self.Meta, "combination_methods", [])
        combined_values: dict[str, dict[str, Any]] = {key: {} for key in combination_methods}
        combined_filters: dict[str, django_filters.Filter] = {}

        for name, value in self.form.cleaned_data.items():
            field_filter: django_filters.Filter = self.filters[name]
            method: str = field_filter._method  # type: ignore[assignment]
            if method in combination_methods:
                combined_values[method][name] = value
                combined_filters[method] = field_filter
                continue

            queryset = field_filter.filter(queryset, value)
            self._verify_that_queryset(queryset)

        for key, values in combined_values.items():
            queryset = combined_filters[key].filter(queryset, values)
            self._verify_that_queryset(queryset)

        return queryset

    @staticmethod
    def _verify_that_queryset(queryset: Any) -> None:
        if not isinstance(queryset, models.QuerySet):  # pragma: no cover
            msg = f"Expected to receive a QuerySet from filters, but got {type(queryset).__name__} instead."
            raise TypeError(msg)



from __future__ import annotations

from .typing import StrEnum

__all__ = [
    "Operation",
]


class Operation(StrEnum):
    # Logical
    AND = "AND"
    OR = "OR"
    XOR = "XOR"
    NOT = "NOT"

    # Comparison single value
    EXACT = "EXACT"
    IEXACT = "IEXACT"
    GT = "GT"
    GTE = "GTE"
    LT = "LT"
    LTE = "LTE"
    CONTAINS = "CONTAINS"
    ICONTAINS = "ICONTAINS"
    STARTSWITH = "STARTSWITH"
    ISTARTSWITH = "ISTARTSWITH"
    ENDSWITH = "ENDSWITH"
    IENDSWITH = "IENDSWITH"
    ISNULL = "ISNULL"
    REGEX = "REGEX"
    IREGEX = "IREGEX"

    # Comparison multiple values
    IN = "IN"
    RANGE = "RANGE"
